# -*- coding: utf-8 -*-
"""
Created on Tue Jul 16 22:13:00 2019
@author: ju-bar

Image convolution kernel routines

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
from numba import jit # include compilation support
import numpy as np
import emilys.optics.aperture as apert
# %%
@jit # compilation decorator, do this, if you intend to call the function several timed
def bwl_gaussian(shape=None, size=0.):
    '''
    Calculates a bandwidth limiting kernel of Gaussian shape with
    given size in pixels. The kernel is returned in the original
    space of the image and is normalized to a total of 1.

    Parameters:
        shape : array (2,)
            shape of the image, number of pixels (rows, columns)
        size : float or array of floats
            size of the bandwidth limit in the original space of the
            image.
            1 float: round gaussian of given width width
                (a = c = size, b = 0)
            2 floats: elliptical gaussian with main axis along the
                    image rows and columns and respective rms widths
                (a = size[0], c = size[1], b = 0)
            3 or more floats: elliptical gaussian with
                (a = size[0], b = size[1], c = size[2])
            
    Return:
        numpy.ndarray, shape=shape, dtype=float
            2D normalized kernel array of requested shape

    Notes:
        Centered bivariate normal distribution,
        see emilys.functions.distributions.bivnorm
        Uses the same function, but simplified code

        p(x,y;a,b,c) = Exp[ -1/2 * (c^2 x^2 + 2 b x y + a^2 y^2) / (a^2 c^2 - b^2)] 
            / (2 Pi sqrt(a^2 c^2 - b^2))

        constraints: abs(a) > 0 && abs(c) > 0 && a^2 c^2 - b^2 != 0
        
        The returned distribution is centered at the [0,0] item of the array.
    '''
    assert np.size(shape)==2, 'expecting parameter 1 as tuple, shape'
    nd = np.array(shape).astype(int)
    assert nd[0] > 0, 'expecting number of rows in first element of parameter 1, shape'
    assert nd[1] > 0, 'expecting number of columns in first element of parameter 1, shape'
    ls = np.array([0.,0.,0.]) # a, b, c
    if (np.size(size)==1):
        ls[0] = np.abs(size)
        ls[2] = np.abs(size)
        assert ls[0] > 0, 'expecting finite size in parameter 2, size'
    elif (np.size(size)==2):
        ls[0] = np.abs(size[0])
        ls[2] = np.abs(size[1])
        assert ls[0] > 0 and ls[2] > 0, 'expecting finite size in parameter 2, size'
    elif (np.size(size)==3):
        ls[0] = np.abs(size[0])
        ls[1] = size[1]
        ls[2] = np.abs(size[2])
        assert ls[0] > 0 and ls[2] > 0, 'expecting finite size in parameter 2, size'
        assert ls[0]**2 * ls[2] ** 2 - ls[1]**2 > 0, \
            '3 size parameters do not define a peak: a^2 c^2 - b^2 <= 0'
    else:
        assert False, 'expecting finite size in parameter 2, size'
    d1 = 1. / (ls[0]**2 * ls[2] ** 2 - ls[1]**2)
    d2 = 0.5 / np.sqrt(d1) / np.pi
    d1 = -0.5 * d1
    kernel = np.zeros(nd, dtype=float)
    nd2 = np.array([int((nd[0]-nd[0]%2)/2), int((nd[1]-nd[1]%2)/2)], dtype=int)
    xi = np.zeros(nd[1], dtype=int)
    for i in range(0, nd[1]):
        xi[i] = (i + nd2[1])%nd[1] - nd2[1]
    norm = 0.
    for j in range(0, nd[0]):
        yj = (j + nd2[0])%nd[0] - nd2[0]
        argy = yj**2 * ls[0]**2
        for i in range(0, nd[1]):
            arg = xi[i]**2 * ls[2]**2 + argy + 2 * xi[i] * yj * ls[1]
            v = d2 * np.exp( d1 * arg )
            norm = norm + v
            kernel[j,i] = v
    kernel = kernel / norm
    return kernel
# %%
def source_delta(kernel):
    '''
    Calculates delta function distribution in a given array.

    Parameters:
        kernel: numpy.ndarray dtype=float
            array receiving distribution values
    
    Return:
        modifies the array kernel
    '''
    kernel[0,0] = 1.
    return 0
# %%
@jit
def source_normal(kernel, samp, src_size):
    '''
    Calculates normal distribution values in a given array.

    Parameters:
        kernel: numpy.ndarray dtype=float
            array receiving distribution values
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates
        src_size: float
            source size (HWHM) in physical units
    
    Return:
        modifies the array kernel
    '''
    srcsz = np.abs(src_size)
    nd = np.array(kernel.shape)
    nd2 = ((nd - nd%2)/2).astype(int)
    krad = 3 * srcsz
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]),nd2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]),nd2[0]))
    kprm = -np.log(2.) / srcsz**2
    norm = 0.
    for j in range(-nky2, nky2+1):
        j1 = j%nd[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i%nd[1]
            rx = samp[0] * i
            r2 = ry2 + rx**2
            if (r2 < kr2):
                kval = np.exp(kprm * r2)
                norm = norm + kval
                kernel[j1,i1] = kval
    kernel[...] = kernel / norm
    return 0
# %%
@jit
def source_cauchy(kernel, samp, src_size):
    '''
    Calculates Cauchy distribution values in a given array.

    Parameters:
        kernel: numpy.ndarray dtype=float
            array receiving distribution values
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates
        src_size: float
            source size (HWHM) in physical units
    
    Return:
        modifies the array kernel
    '''
    srcsz = np.abs(src_size)
    nd = np.array(kernel.shape)
    nd2 = ((nd - nd%2)/2).astype(int)
    krad = 10 * srcsz
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]),nd2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]),nd2[0]))
    rprm = 1.3047660265 * srcsz
    afac = 0.5 * rprm / np.pi
    kprm = rprm**2
    kap = 1.5
    norm = 0.
    for j in range(-nky2, nky2+1):
        j1 = j%nd[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i%nd[1]
            rx = samp[0] * i
            r2 = ry2 + rx**2
            if (r2 < kr2):
                kval = afac / (r2 + kprm)**kap
                norm = norm + kval
                kernel[j1,i1] = kval
    kernel[...] = kernel / norm
    return 0
# %%
@jit
def source_disk(kernel, samp, src_size):
    '''
    Calculates disk distribution values in a given array.

    Parameters:
        kernel: numpy.ndarray dtype=float
            array receiving distribution values
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates
        src_size: float
            source size in physical units
    
    Return:
        modifies the array kernel
    '''
    srcsz = np.abs(src_size)
    nd = np.array(kernel.shape)
    nd2 = ((nd - nd%2)/2).astype(int)
    krad = 1.5 * srcsz
    nkx2 = int(min(np.ceil(krad / samp[0]),nd2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]),nd2[0]))
    kprm = srcsz
    norm = 0.
    for j in range(-nky2, nky2+1):
        j1 = j%nd[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i%nd[1]
            rx = samp[0] * i
            rm = np.sqrt(ry2 + rx**2)
            if (rm < krad):
                kval =  0.5 - 0.5 * np.tanh((rm/kprm - 1.0) * 100.0)
                norm = norm + kval
                kernel[j1,i1] = kval
    kernel[...] = kernel / norm
    return 0
# %%
def source_distribution(shape, samp, src_size, src_type=1):
    '''
    Calculates a normalized source distribution function for convolution
    of incoherent images.

    Parameters:
        shape : array, int, (2,)
            shape of the image, number of pixels (rows, columns)
        samp : array, flat (2,)
            sampling rates along image rows and columns (x,y) in
            physical units
        src_size : float
            source size (HWHM) in physical units
        src_type : int
            distribution type
            0 = delta function = size ignored
            1 = normal distribution
            2 = Cauchy (Lorentzian) distribution
            3 = Disk (Sigmoidal) distribution

    Return:
        numpy.ndarray, shape=shape, dtype=float
            2D normalized kernel array of requested shape
    
    Notes:
        This function implements round source distributions only.
        The returned distribution is centered at the [0,0] item of the array.
    '''
    assert np.size(shape)==2, 'expecting parameter 1 as tuple, shape'
    nd = np.array(shape).astype(int)
    assert nd[0] > 0, 'expecting number of rows in first element of parameter 1, shape'
    assert nd[1] > 0, 'expecting number of columns in first element of parameter 1, shape'
    ntuse = int(src_type)
    srcsz = np.abs(src_size)
    if (srcsz == 0.): ntuse = 0
    assert ntuse >= 0 and ntuse <=3, 'expecting integer parameter src_type in range 0 ... 3'
    kernel = np.zeros(nd, dtype=float)
    if (ntuse == 0):
        source_delta(kernel)
    elif (ntuse == 1):
        source_normal(kernel, samp, srcsz)
    elif (ntuse == 2):
        source_cauchy(kernel, samp, srcsz)
    elif (ntuse == 3):
        source_disk(kernel, samp, srcsz)
    return kernel

# %%
@jit
def filter_normal_iso_2d(ndim, samp, flt_size):
    '''
    Returns a Gaussian isotropic filter kernel for a low-pass in an array of given dimension.

    Parameters:
        ndim: numpy.ndarray shape=(2,) dtype=int
            kernel dimensions (nrows, ncolumns), (ny, nx)
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates (sx, sy)
        flt_size: float
            filter kernel size (HWHM) in physical units
    
    Return:
        kernel: numpy.ndarray shape=(nrows, ncolumns), dtype=float
    '''
    srcsz = np.abs(flt_size)
    kernel = np.zeros((ndim[0], ndim[1]), dtype=float)
    ndim2 = np.array([ndim[0] >> 1, ndim[1] >> 1])
    krad = 4 * srcsz # calculate filter kernel up to 4 sigma
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]),ndim2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]),ndim2[0]))
    kprm = -np.log(2.) / srcsz**2
    for j in range(-nky2, nky2+1):
        j1 = j % ndim[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i % ndim[1]
            rx = samp[0] * i
            r2 = ry2 + rx**2
            if (r2 < kr2):
                kval = np.exp(kprm * r2)
                kernel[j1,i1] = kval
    return kernel
# %%
@jit
def filter_normal_2d(ndim, samp, flt_size_a, flt_size_b, flt_dir):
    '''
    Returns a Gaussian anisotropic filter kernel for a low-pass in an array of given dimension.
    The filter is returned with value 1 at pixel [0,0]. Normalization should be done numerically.
    The analytical norm if treated as distribution is np.pi * flt_size_a * flt_size_b / np.log(2.).

    Parameters:
        ndim: numpy.ndarray shape=(2,) dtype=int
            kernel dimensions (nrows, ncolumns), (ny, nx)
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates (sx, sy)
        flt_size_a: float
            filter kernel size (HWHM) in physical units along orientation a
        flt_size_b: float
            filter kernel size (HWHM) in physical units along orientation b perpendicular to a
        flt_dir: float
            filter kernel a axis orientation to the grid row axis in radians
    
    Return:
        kernel: numpy.ndarray shape=(nrows, ncolumns), dtype=float
    '''
    sz_a = np.abs(flt_size_a)
    sz_b = np.abs(flt_size_b)
    sz_d = flt_dir
    kernel = np.zeros((ndim[0], ndim[1]), dtype=float)
    ndim2 = np.array([ndim[0] >> 1, ndim[1] >> 1])
    krad = 4 * np.amax([sz_a, sz_b]) # calculate filter kernel up to 4 sigma
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]), ndim2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]), ndim2[0]))
    kprm = -0.5 * np.log(2.) / (sz_a**2 * sz_b**2)
    vprmp = sz_a**2 + sz_b**2
    vprmn = sz_a**2 - sz_b**2
    sdir2 = np.sin(2. * sz_d)
    cdir2 = np.cos(2. * sz_d)
    for j in range(-nky2, nky2+1):
        j1 = j % ndim[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i % ndim[1]
            rx = samp[0] * i
            rx2 = rx * rx
            r2 = ry2 + rx2
            if (r2 < kr2):
                kval = np.exp(kprm * (vprmp * r2 - vprmn * ((rx2 - ry2) * cdir2 + 2. * rx * ry * sdir2)))
                kernel[j1,i1] = kval
    return kernel
# %%
@jit
def filter_exp4_2d_iso(ndim, samp, flt_size):
    '''
    Returns an exponential g^4 isotropic filter kernel for a low-pass in an array of given dimension.

    Parameters:
        ndim: numpy.ndarray shape=(2,) dtype=int
            kernel dimensions (nrows, ncolumns), (ny, nx)
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates (sx, sy)
        flt_size: float
            filter kernel size (HWHM) in physical units
    
    Return:
        kernel: numpy.ndarray shape=(nrows, ncolumns), dtype=float
    '''
    srcsz = np.abs(flt_size)
    kernel = np.zeros((ndim[0], ndim[1]), dtype=float)
    ndim2 = np.array([ndim[0] >> 1, ndim[1] >> 1])
    krad = 4 * srcsz # calculate filter kernel up to 4 sigma
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]),ndim2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]),ndim2[0]))
    kprm = -np.log(2.) / srcsz**4
    for j in range(-nky2, nky2+1):
        j1 = j % ndim[0]
        ry = samp[1] * j
        ry2 = ry**2
        for i in range(-nkx2, nkx2+1):
            i1 = i % ndim[1]
            rx = samp[0] * i
            r2 = ry2 + rx**2
            if (r2 < kr2):
                kval = np.exp(kprm * r2**2)
                kernel[j1,i1] = kval
    return kernel
# %%
@jit
def filter_exp4_2d(ndim, samp, flt_size_a, flt_size_b, flt_dir):
    '''
    Returns a exponential forth-order anisotropic filter kernel for a low-pass in an array of
    given dimension. The filter is returned with value 1 at pixel [0,0]. Normalization should
    be done numerically.

    Parameters:
        ndim: numpy.ndarray shape=(2,) dtype=int
            kernel dimensions (nrows, ncolumns), (ny, nx)
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates (sx, sy)
        flt_size_a: float
            filter kernel size (HWHM) in physical units along orientation a
        flt_size_b: float
            filter kernel size (HWHM) in physical units along orientation b perpendicular to a
        flt_dir: float
            filter kernel a axis orientation to the grid row axis in radians
    
    Return:
        kernel: numpy.ndarray shape=(nrows, ncolumns), dtype=float
    '''
    sz_a = np.abs(flt_size_a)
    sz_b = np.abs(flt_size_b)
    sz_d = flt_dir
    kernel = np.zeros((ndim[0], ndim[1]), dtype=float)
    ndim2 = np.array([ndim[0] >> 1, ndim[1] >> 1])
    krad = 3 * np.amax([sz_a, sz_b]) # calculate filter kernel up to 3 sigma
    kr2 = krad**2
    nkx2 = int(min(np.ceil(krad / samp[0]), ndim2[1]))
    nky2 = int(min(np.ceil(krad / samp[1]), ndim2[0]))
    kprm = -np.log(2.)
    sdir = np.sin(sz_d)
    cdir = np.cos(sz_d)
    for j in range(-nky2, nky2+1):
        j1 = j % ndim[0]
        ry = samp[1] * j
        ry2 = ry**2
        py2prm = ry2 * (cdir**2 / sz_b**2 + sdir**2 / sz_a**2)
        for i in range(-nkx2, nkx2+1):
            i1 = i % ndim[1]
            rx = samp[0] * i
            rx2 = rx * rx
            px2prm = rx2 * (cdir**2 / sz_a**2 + sdir**2 / sz_b**2)
            pxyprm = 2. * rx * ry * (1./sz_a**2 - 1./sz_b**2) * sdir * cdir
            r2 = ry2 + rx2
            if (r2 < kr2):
                kval = np.exp( kprm * (px2prm + py2prm + pxyprm)**2 )
                kernel[j1,i1] = kval
    return kernel
# %%
@jit
def filter_sigmoid_2d(ndim, samp, flt_size, smoothness=0., anisotropy=0., orientation=0.):
    '''
    Returns a sigmodial anisotropic filter kernel for a low-pass in an array of given dimension.

    Parameters:
        ndim: numpy.ndarray shape=(2,) dtype=int
            kernel dimensions (nrows, ncolumns), (ny, nx)
        samp: numpy.ndarray shape=(2,) dtype=float
            row and column sampling rates (sx, sy)
        flt_size: float
            filter size (HWHM) in physical units
        smoothness: float
            edge smoothness of the kernel in physical units
        anisotropy: float
            kernel anisotropy relative to flt_size
        orientation: float
            anisotropy orientation of large axis to rows, ]-pi, pi]
    
    Return:
        kernel: numpy.ndarray shape=(nrows, ncolumns), dtype=float
    '''
    srcsz = np.abs(flt_size)
    kernel = np.zeros((ndim[0], ndim[1]), dtype=float)
    ndim2 = np.array([ndim[0] >> 1, ndim[1] >> 1])
    apert.aperture_a_grid(kernel,ndim2.astype(float),np.array([[samp[0],0.],[0.,samp[1]]]),np.array([0.,0.]),srcsz,smoothness,anisotropy,orientation)
    kernel[...] = np.roll(kernel, shift = ndim2, axis = (0,1))
    return kernel
# %%
def normal_1d(n, p0, sigma):
    """

    Returns a 1d array of length n with a normal distribution
    of r.m.s. width sigma centered a position p0.

    Parameters
    ----------
        n : int
            number of samples of the 1d output
        p0 : float
            center position of the normal distribution
        sigma : float
            r.m.s. width of the distribution

    Returns
    -------
        numpy.ndarray of length n

    Remarks
    -------
        periodic wrap around is assumed

    """
    assert n > 1, 'n must be larger than one'
    assert np.abs(sigma) > 0, 'sigma must be non-zero'
    pf = -0.5 / sigma**2
    n2 = n>>1 # index of the center pixel
    l_kern = np.zeros(n)
    n0 = int(np.floor(p0))
    p0_frac = p0 - n0
    sum_kern = 0.
    for i in range(0, n):
        x = 1. * (i - n2) + p0_frac
        v_kern = np.exp(pf * x**2)
        l_kern[i] = v_kern
        sum_kern += v_kern
    assert sum_kern > 0, 'intermediate kernel power is zero'
    return np.roll(l_kern / sum_kern, n0-n2)


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 14:42:00 2019
@author: ju-bar

Array data plotting wrapper

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
import matplotlib.pyplot as plt # include mathematic plotting
import numpy as np # include numeric functions
# %%
def get_value_range(array, vrange=np.array([0.,1.]), vrangetype='default'):
    """
    Calculates the range of values of an array according to range type
    and range limit parameters. May return array data with modified
    values.

    Parameters
    ----------
    array : array of 2 dimensions
        array data intended for plotting
    vrange : array shape(2,)
        range limits
    vrangetype : string
        value range setting mode:
            'default' : min. to max. of array values,
                        modified relatively by vrange
            'direct'  : plots from vrange[0] to vrange[1]
            'modulo'  : plots from vrange[0] to vrange[1],
                        modifies array -> arrshow with by clipping
                        all values into this range using a modulo
                        operation
            'zerosym' : symmetric around zero covering max. absolute
                        array value, modified relatively by vrange

    Return
    ------
    array
        [rmin, rmax, arrshow]
        rmin : float
            minimum value to be plotted
        rmax : float
            maximum value to be plotted
        arrshow : array
            modified values to be plotted in range [rmin, rmax]
    """
    vmin = np.amin(array)
    vmax = np.amax(array)
    rmin = vmin + vrange[0] * (vmax - vmin)
    rmax = vmin + vrange[1] * (vmax - vmin)
    arrshow = array
    if vrangetype == 'zerosym': # set value range symmetric around zero
        if vmax < 0:
            vmax = np.abs(vmin)
        else:
            vmin = -vmax
        rmin = vmin + vrange[0] * (vmax - vmin)
        rmax = vmin + vrange[1] * (vmax - vmin)
    elif vrangetype == 'direct': # set the value range directly from vrange
        rmin = vrange[0]
        rmax = vrange[1]
    elif vrangetype == 'modulo': # use modulo on the value range
        rmin = vrange[0]
        rmax = vrange[1]
        arrshow = np.mod(array - vrange[0], vrange[1] -vrange[0]) + vrange[0]
    return [rmin, rmax, arrshow]
# %%
def arrayplot2d(array, pixscale=1, colscale='gray', dpi=72, 
                vrange=np.array([0.,1.]), vrangetype='default'):
    """
    Uses matplotlib.pyplot to plot a pixel-true image of the input 'array'.

    Parameters
    ----------
    array : array shape(n,m)
        input data to plot as image
    pixscale : int or float : default 1
        number of image dots per array element
    colscale : string : default 'gray'
        color palette identifier
    dpi : int : default 72
        dots per inch for image output
    vrange : array shape(2,) : default [0.,1.]
        value range limits
    vrangetype : string : default 'default'
        value range setting mode:
            'default' : min. to max. of array values,
                        modified relatively by vrange
            'direct'  : plots from vrange[0] to vrange[1]
            'modulo'  : plots from vrange[0] to vrange[1],
                        modifies array -> arrshow with by clipping
                        all values into this range using a modulo
                        operation
            'zerosym' : symmetric around zero covering max. absolute
                        array value, modified relatively by vrange
    Return
    ------
    [matplotlib.pyplot.figure, matplotlib.pyplot.axes]
        figure and axes object used for plotting the array

    """
    nd = array.shape
    fig = plt.figure(figsize=(pixscale*nd[1]/dpi,pixscale*nd[0]/dpi), dpi=dpi)
    ax = fig.add_axes([0,0,1,1], frameon=False, aspect=1)
    ax.axis('off')
    rmin, rmax, arrshow = get_value_range(array, vrange, vrangetype)
    ax.imshow(arrshow, cmap = colscale, origin = 'lower', vmin = rmin,
              vmax = rmax)
    return [fig,ax]

#%%



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 16 12:04:00 2019
@author: ju-bar

Calculation of 2-dimensional apertures

This code is part of the 'emilys' repository
https://github.com/ju-bar/emilys
published under the GNU General Publishing License, version 3

"""
# %%
from numba import jit # include compilation support
import numpy as np # include numeric functions
# %%
@jit
def aperture(q=np.array([0.,0.]), q0=np.array([0.,0.]), qlim=1., qsmt=0.):
    """

    Calculates transmission values for a round smooth aperture in 2D.
    (numba compiled)

    Parameters:
        q : np.array[shape=(2)]
            query point
        q0 : np.array[shape=(2)]
            aperture center position
        qlim : float
            aperture radius
        qsmt : float
            edge smoothness
    """
    qs = np.abs(qsmt)
    ql = np.abs(qlim)
    if ql > 0.: # the aperture is open
        dq = q - q0
        dqm2 = np.dot(dq,dq)
        if dqm2 > 0.: # the query point is away from the aperture center
            dqe = dq / np.sqrt(dqm2) # unit vector pointing from aperture center to query point
            dql = dqe * ql # vector from aperture center to its edge in direction of query point, if the aperture is not round, this would need to be modified
            if qs > 0.: # the edge is smooth
                dqd = dq - dql # vector from aperture edge to query point (along the connection to the aperture center)
                arg = np.pi*np.dot(dqd,dqe) / qs # distance from the edge rescaled to Pi/qsmt
                return 0.5*(1. - np.tanh(arg)) # sigmoid edge transition from 1 to 0 
            else: # the edge is sharp
                if np.dot(dql,dql) > dqm2: # the query point is inside the aperture radius
                    return 1.
                else: # the query point is outside the aperture radius
                    return 0.
        else: # the query point is in the aperture center
            return 1.
    else: # the aperture is closed
        return 0.
    return 0. # default exit
# %%
@jit
def aperture_a(q=np.array([0.,0.]), q0=np.array([0.,0.]), qlim=1., qsmt=0., qa=0., qp=0.):
    """

    Calculates transmission values for a smooth aperture in 2D with 2-fold edge distortion
    (numba compiled)

    Parameters:
        q : np.array[shape=(2)]
            query point
        q0 : np.array[shape=(2)]
            aperture center position
        qlim : float
            aperture radius
        qsmt : float
            edge smoothness
        qa : float
            magnitude of 2-fold distortion
        qp : float
            direction of positive axis of 2-fold dist. [-Pi, Pi[
            
    """
    if (qlim > 0.):
        a1x = qa * np.cos( 2.* qp ) / qlim # rel. 2-fold distortion x component
        a1y = qa * np.sin( 2.* qp ) / qlim # rel. 2-fold distortion y component
        adet = 1. + (qa/qlim)**2
        if (adet == 0.0):
            return 0.0 # closed aperture by line distortion, exit
        radet = 1. / adet
        dq = q - q0 # distance vector to center
        dqx1 = ((1. - a1x)*dq[0] - a1y*dq[1])*radet # back project dqx to undistorted plane
        dqy1 = (-a1y*dq[0] + (1. + a1x)*dq[1])*radet # back project dqy to undistorted plane
        dq2 = dqx1 * dqx1 + dqy1 * dqy1
        dqm = np.sqrt(dq2) # undistorted distance from aperture center
        dqs = abs(qsmt) # absolute smoothness
        if (dqm > 0.0): # handle default case (beam is somewhere in aperture area)
            if (dqs > 0.0): # calculate smoothed aperture
                darg = np.pi * (dqm - qlim) / dqs
                return (1. - np.tanh(darg))*0.5 # aperture value
            else: # sharp aperture
                if (dqm < qlim):
                    return 1.0 # point is in the aperture
        else: # handle on-axis case
            return 1.0 # ... always transmit this beam
    else:
        return 0.
    return 0.

# %%
@jit
def aperture_dist3(q=np.array([0.,0.]), q0=np.array([0.,0.]), qlim=1., qsmt=0., qdist=np.array([])):
    """

    Calculates transmission values for a smooth aperture in 2D with edge distortions up to 3rd order.
    (numba compiled)

    Parameters:
        q : np.array[shape=(2)]
            query point
        q0 : np.array[shape=(2)]
            aperture center position
        qlim : float
            aperture radius
        qsmt : float
            edge smoothness
        qdist : np.array[shape=(na)]
            list of edge distortion parameters, length = na
            0,1 : two-fold distortion x, y
            2,3 : three-fold distortion x, y
            4,5 : four-fold distortion x, y
    """
    na = qdist.size # number of distortion parameters
    #print("dbg aperture_dist3: distortion parameters: na = ",na)
    qs = np.abs(qsmt) # strength of edge smoothness
    ql = np.abs(qlim) # aperture radius (must be positive)
    if ql > 0.: # the aperture is open
        dq = q - q0 # vector from aperture center to query point
        dqm2 = np.dot(dq,dq) # squared distance of query point from center
        if dqm2 > 0.: # the query point is away from the aperture center
            dqe = dq / np.sqrt(dqm2) # unit vector pointing from aperture center to query point [cos(p),sin(p)]
            dql = dqe * ql # vector from aperture center to its edge in direction of query point
            dql3 = dql
            if na > 0: # there are edge distortion parameters
                qx1 = dqe[0] # cos(p)
                qy1 = dqe[1] # sin(p)
                qx2 = qx1*qx1 # cos(p)**2
                qx3 = qx2*qx1 # cos(p)**3
                qy2 = qy1*qy1 # sin(p)**2
                qy3 = qy2*qy1 # sin(p)**3
                lbd3x = np.array([qx1,qy1,qx2-qy2,-2*qx1*qy1,qx3-3*qx1*qy2,3*qx2*qy1-qy3]) # aberration x basis
                lbd3y = np.array([-qy1,qx1,-2*qx1*qy1,qy2-qx2,qy3-3*qx2*qy1,qx3-3*qx1*qy2]) # aberration y basis
                for l in range(0, na):
                    #print("dbg aperture_dist3: l=",l,", qdist[l]=",qdist[l])
                    #print("dbg aperture_dist3: bqx(l)=",lbd3x[l],", bqy(l)=",lbd3y[l])
                    dql3 = dql3 + np.array([lbd3x[l]*qdist[l],lbd3y[l]*qdist[l]])
            if qs > 0.: # the edge is smooth
                dqd = dq - dql3 # vector from aperture edge to query point (along the connection to the aperture center)
                arg = np.pi*np.dot(dqd,dqe) / qs # distance from the edge rescaled to Pi/qsmt
                return 0.5*(1. - np.tanh(arg)) # sigmoid edge transition from 1 to 0 
            else: # the edge is sharp
                if np.dot(dql3,dql3) > dqm2: # the query point is inside the aperture radius
                    return 1.
                else: # the query point is outside the aperture radius
                    return 0.
        else: # the query point is in the aperture center
            return 1.
    else: # the aperture is closed
        return 0.
    return 0. # default exit
# %%
@jit
def aperture_grid(arr, p0=np.array([0.,0.]), sq=np.array([[1.,0.],[0.,1.]]),
                  q0=np.array([0.,0.]), qlim=1., qsmt=0.):
    """

    Calculates transmission values for a round smooth aperture on a 2D grid.
    (numba compiled)

    Parameters:
        arr : np.array[shape=(ny,nx)]
            array receiving aperture transmission values
        p0 : np.array[shape=(2)]
            origin pixel
        sq : np.array[shape=(2,2)]
            physical scale
        q0 : np.array[shape=(2)]
            aperture center position on physical scale
        qlim : float
            aperture radius on physical scale
        qsmt : float
            edge smoothness on physical scale
    """
    nd = arr.shape
    #print("dbg aperture_grid: nd = " % nd)
    for j in range(0, nd[1]):
        for i in range(0, nd[0]):
            dp = np.array([i,j]) - p0
            q = np.dot(sq,dp)
            arr[j,i] = aperture(q, q0, qlim, qsmt)
    return 0
# %%
@jit
def aperture_a_grid(arr, p0=np.array([0.,0.]),
                    sq=np.array([[1.,0.],[0.,1.]]),
                    q0=np.array([0.,0.]),
                    qlim=1., qsmt=0., qa=0., qp=0.):
    """

    Calculates transmission values for a smooth aperture with 
    2-fold edge distortion on a 2D grid.
    (numba compiled)

    Parameters:
        arr : np.array[shape=(ny,nx)]
            array receiving aperture transmission values
        p0 : np.array[shape=(2)]
            origin pixel
        sq : np.array[shape=(2,2)]
            physical scale
        q0 : np.array[shape=(2)]
            aperture center position on physical scale
        qlim : float
            aperture radius on physical scale
        qsmt : float
            edge smoothness on physical scale
        qa : float
            magnitude of 2-fold distortion on physical scale
        qp : float
            direction of positive axis of 2-fold dist. [-Pi, Pi[
    """
    nd = arr.shape
    for j in range(0, nd[1]):
        for i in range(0, nd[0]):
            dp = np.array([i,j]) - p0
            q = np.dot(sq,dp)
            arr[j,i] = aperture_a(q, q0, qlim, qsmt, qa, qp)
    return 0
# %%
@jit
def aperture_dist3_grid(arr, p0=np.array([0.,0.]),
                        sq=np.array([[1.,0.],[0.,1.]]),
                        q0=np.array([0.,0.]), qlim=1., qsmt=0.,
                        qdist=np.array([])):
    """

    Calculates transmission values for a smooth aperture with edge
    distortions up to 3rd order on a 2D grid.
    (numba compiled)

    Parameters:
        arr : np.array[shape=(ny,nx)]
            array receiving aperture transmission values
        p0 : np.array[shape=(2)]
            origin pixel
        sq : np.array[shape=(2,2)]
            physical scale
        q0 : np.array[shape=(2)]
            aperture center position on physical scale
        qlim : float
            aperture radius on physical scale
        qsmt : float
            edge smoothness on physical scale
        qdist : np.array[shape=(na)]
            list of edge distortion parameters, length = na
            0,1 : two-fold distortion x, y
            2,3 : three-fold distortion x, y
            4,5 : four-fold distortion x, y
    """
    nd = arr.shape
    #print("dbg aperture_dist3_grid: nd = ", nd)
    #print("dbg aperture_dist3_grid: na = ", qdist.size)
    for j in range(0, nd[1]):
        for i in range(0, nd[0]):
            dp = np.array([i,j]) - p0
            q = np.dot(sq,dp)
            arr[j,i] = aperture_dist3(q, q0, qlim, qsmt, qdist)
    return 0
#%%
@jit
def hann(n):
    '''
    Returns a 1d preset with a Hann window.
    '''
    return np.sin(np.pi*np.arange(0,n)/n)**2

